#ifndef HASH_OPEN_MAP_HPP_
#define HASH_OPEN_MAP_HPP_

#include <string>
#include <iostream>
#include <sstream>
#include <initializer_list>
#include "ics_exceptions.hpp"
#include "pair.hpp"


namespace ics {


#ifndef undefinedhashdefined
#define undefinedhashdefined
template<class T>
int undefinedhash (const T& a) {return 0;}
#endif /* undefinedhashdefined */

//Instantiate the templated class supplying thash(a): produces a hash value for a.
//If thash is defaulted to undefinedhash in the template, then a constructor must supply chash.
//If both thash and chash are supplied, then they must be the same (by ==) function.
//If neither is supplied, or both are supplied but different, TemplateFunctionError is raised.
//The (unique) non-undefinedhash value supplied by thash/chash is stored in the instance variable hash.
template<class KEY,class T, int (*thash)(const KEY& a) = undefinedhash<KEY>> class HashOpenMap {
  public:
    typedef ics::pair<KEY,T>      Entry;         //Stores key and its associated value
    typedef ics::pair<int,Entry>  HashedEntry;   //Stores hash(key) and Entry (key/value); see the h_entry member
    typedef int (*hashfunc) (const KEY& a);

    //Destructor/Constructors
    ~HashOpenMap ();

    HashOpenMap          (double the_load_threshold = 1.0, int (*chash)(const KEY& a) = undefinedhash<KEY>);
    explicit HashOpenMap (int initial_bins, double the_load_threshold = 1.0, int (*chash)(const KEY& k) = undefinedhash<KEY>);
    HashOpenMap          (const HashOpenMap<KEY,T,thash>& to_copy, double the_load_threshold = 1.0, int (*chash)(const KEY& a) = undefinedhash<KEY>);
    explicit HashOpenMap (const std::initializer_list<Entry>& il, double the_load_threshold = 1.0, int (*chash)(const KEY& a) = undefinedhash<KEY>);

    //Iterable class must support "for-each" loop: .begin()/.end() and prefix ++ on returned result
    template <class Iterable>
    explicit HashOpenMap (const Iterable& i, double the_load_threshold = 1.0, int (*chash)(const KEY& a) = undefinedhash<KEY>);


    //Queries
    bool empty      () const;
    int  size       () const;
    bool has_key    (const KEY& key) const;
    bool has_value  (const T& value) const;
    std::string str () const; //supplies useful debugging information; contrast to operator <<


    //Commands
    T    put   (const KEY& key, const T& value);
    T    erase (const KEY& key);
    void clear ();

    //Iterable class must support "for-each" loop: .begin()/.end() and prefix ++ on returned result
    template <class Iterable>
    int put_all(const Iterable& i);


    //Operators

    T&       operator [] (const KEY&);
    const T& operator [] (const KEY&) const;
    HashOpenMap<KEY,T,thash>& operator = (const HashOpenMap<KEY,T,thash>& rhs);
    bool operator == (const HashOpenMap<KEY,T,thash>& rhs) const;
    bool operator != (const HashOpenMap<KEY,T,thash>& rhs) const;

    template<class KEY2,class T2, int (*hash2)(const KEY2& a)>
    friend std::ostream& operator << (std::ostream& outs, const HashOpenMap<KEY2,T2,hash2>& m);



  public:
    class Iterator {
      public:
        //Private constructor called in begin/end, which are friends of HashOpenMap<T>
        ~Iterator();
        Entry       erase();
        std::string str  () const;
        HashOpenMap<KEY,T,thash>::Iterator& operator ++ ();
        HashOpenMap<KEY,T,thash>::Iterator  operator ++ (int);
        bool operator == (const HashOpenMap<KEY,T,thash>::Iterator& rhs) const;
        bool operator != (const HashOpenMap<KEY,T,thash>::Iterator& rhs) const;
        Entry& operator *  () const;
        Entry* operator -> () const;
        friend std::ostream& operator << (std::ostream& outs, const HashOpenMap<KEY,T,thash>::Iterator& i) {
          outs << i.str(); //Use the same meaning as the debugging .str() method
          return outs;
        }
        friend Iterator HashOpenMap<KEY,T,thash>::begin () const;
        friend Iterator HashOpenMap<KEY,T,thash>::end   () const;

      private:
        //If can_erase is false, must ++ to reach next value
        int                       current; //h_entry Index
        HashOpenMap<KEY,T,thash>* ref_map;
        int                       expected_mod_count;
        bool                      can_erase = true;

        //Called in friends begin/end
        Iterator(HashOpenMap<KEY,T,thash>* iterate_over, int initial);
    };


    Iterator begin () const;
    Iterator end   () const;


private:
  int (*hash)(const KEY& k);          //Hashing function used (from template or constructor)

  const static int bin_empty  = -1;   //Special values that can be stored in the bin array,
  const static int was_occupied = -2; //  along with non-negative indexes into the h_entry array
  int*         bin     = nullptr;     //Non-contiguous int array: bin[i] is either -1, -2, or an index into the h_entry array
  HashedEntry* h_entry = nullptr;     //Contiguous HashedEntry array storing hash(key) and key/value as pair of pair
                                      //bin/h_entry are allocated in constructors and in ensure_lengths
  double load_threshold;              //used/bins must always be <= load_threshold (see ensure_lengths)
  int bins             = 1;           //Allocated length of bin array (>= 1, so hash_compress doesn't % 0)
  int entries          = 1;           //Allocated length of h_entry array (can be < bins, if load_factor < 1)
  int used             = 0;           //# of key/value associations in this map; used <= entries (see ensure_lengths)
                                      //  bins-used = # of bin_empty/was_occupied values in bin array
  int mod_count        = 0;           //For sensing concurrent modification


  //Helper methods
  //Given key, compute/return its hashed value and which bin it compresses to (%bins in range [0,bins-1])
  void hash_compress   (const KEY& key, int& hashed, int& bin_index)  const;
  int  find_key        (const KEY& key)  const;  //Returns index to key's bin or -1 if  key is not in any bin
  int  next_unoccupied (int bin_index)   const;  //this bin_index (if unoccupied) or next bin_index beyond that
  //   is unoccupied; advances using linear probing
  void ensure_lengths  (int new_used);           //Reallocate bin/h_entry if load factor > load_threshold/more entries needed
};




////////////////////////////////////////////////////////////////////////////////
//
//HashOpenMap class and related definitions

//Destructor/Constructors

template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::~HashOpenMap() {
}


template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::HashOpenMap(double the_load_threshold, int (*chash)(const KEY& k))
{
}


template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::HashOpenMap(int initial_bins, double the_load_threshold, int (*chash)(const KEY& k))
{
}


template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::HashOpenMap(const HashOpenMap<KEY,T,thash>& to_copy, double the_load_threshold, int (*chash)(const KEY& a))
{
}


template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::HashOpenMap(const std::initializer_list<Entry>& il, double the_load_threshold, int (*chash)(const KEY& k))
{
}


template<class KEY,class T, int (*thash)(const KEY& a)>
template <class Iterable>
HashOpenMap<KEY,T,thash>::HashOpenMap(const Iterable& i, double the_load_threshold, int (*chash)(const KEY& k))
{
}


////////////////////////////////////////////////////////////////////////////////
//
//Queries

template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::empty() const {return false;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
int HashOpenMap<KEY,T,thash>::size() const {return 0;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::has_key (const KEY& key) const {return false;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::has_value (const T& value) const {return false;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
std::string HashOpenMap<KEY,T,thash>::str() const {return "";
}


////////////////////////////////////////////////////////////////////////////////
//
//Commands

template<class KEY,class T, int (*thash)(const KEY& a)>
T HashOpenMap<KEY,T,thash>::put(const KEY& key, const T& value) {return value;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
T HashOpenMap<KEY,T,thash>::erase(const KEY& key) {return h_entry[0].second.second;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
void HashOpenMap<KEY,T,thash>::clear() {
}


template<class KEY,class T, int (*thash)(const KEY& a)>
template<class Iterable>
int HashOpenMap<KEY,T,thash>::put_all(const Iterable& i) {return 0;
}


////////////////////////////////////////////////////////////////////////////////
//
//Operators

template<class KEY,class T, int (*thash)(const KEY& a)>
T& HashOpenMap<KEY,T,thash>::operator [] (const KEY& key) {return h_entry[0].second.second;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
const T& HashOpenMap<KEY,T,thash>::operator [] (const KEY& key) const {return h_entry[0].second.second;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>& HashOpenMap<KEY,T,thash>::operator = (const HashOpenMap<KEY,T,thash>& rhs) {return *this;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::operator == (const HashOpenMap<KEY,T,thash>& rhs) const {return false;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::operator != (const HashOpenMap<KEY,T,thash>& rhs) const {return false;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
std::ostream& operator << (std::ostream& outs, const HashOpenMap<KEY,T,thash>& m) {return outs;
}


////////////////////////////////////////////////////////////////////////////////
//
//Iterator constructors

template<class KEY,class T, int (*thash)(const KEY& a)>
auto HashOpenMap<KEY,T,thash>::begin () const -> HashOpenMap<KEY,T,thash>::Iterator {return Iterator(const_cast<HashOpenMap<KEY,T,thash>*>(this),0);
}


template<class KEY,class T, int (*thash)(const KEY& a)>
auto HashOpenMap<KEY,T,thash>::end () const -> HashOpenMap<KEY,T,thash>::Iterator {return Iterator(const_cast<HashOpenMap<KEY,T,thash>*>(this),0);
}


////////////////////////////////////////////////////////////////////////////////
//
//Private helper methods

template<class KEY,class T, int (*thash)(const KEY& a)>
void HashOpenMap<KEY,T,thash>::hash_compress (const KEY& key, int& hashed, int& bin_index) const {
}


template<class KEY,class T, int (*thash)(const KEY& a)>
int HashOpenMap<KEY,T,thash>::find_key (const KEY& key) const {
}


template<class KEY,class T, int (*thash)(const KEY& a)>
int HashOpenMap<KEY,T,thash>::next_unoccupied (int bin_index) const {
}


template<class KEY,class T, int (*thash)(const KEY& a)>
void HashOpenMap<KEY,T,thash>::ensure_lengths(int new_used) {
}






////////////////////////////////////////////////////////////////////////////////
//
//Iterator class definitions

template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::Iterator::Iterator(HashOpenMap<KEY,T,thash>* iterate_over, int initial)
{
}


template<class KEY,class T, int (*thash)(const KEY& a)>
HashOpenMap<KEY,T,thash>::Iterator::~Iterator()
{}


template<class KEY,class T, int (*thash)(const KEY& a)>
auto HashOpenMap<KEY,T,thash>::Iterator::erase() -> Entry {return ref_map->h_entry[0].second;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
std::string HashOpenMap<KEY,T,thash>::Iterator::str() const {return "";
}


template<class KEY,class T, int (*thash)(const KEY& a)>
auto  HashOpenMap<KEY,T,thash>::Iterator::operator ++ () -> HashOpenMap<KEY,T,thash>::Iterator& {return *this;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
auto  HashOpenMap<KEY,T,thash>::Iterator::operator ++ (int) -> HashOpenMap<KEY,T,thash>::Iterator {return *this;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::Iterator::operator == (const HashOpenMap<KEY,T,thash>::Iterator& rhs) const {
}


template<class KEY,class T, int (*thash)(const KEY& a)>
bool HashOpenMap<KEY,T,thash>::Iterator::operator != (const HashOpenMap<KEY,T,thash>::Iterator& rhs) const {return false;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
pair<KEY,T>& HashOpenMap<KEY,T,thash>::Iterator::operator *() const {return ref_map->h_entry[0].second;
}


template<class KEY,class T, int (*thash)(const KEY& a)>
pair<KEY,T>* HashOpenMap<KEY,T,thash>::Iterator::operator ->() const {return &(ref_map->h_entry[0].second);
}


}

#endif /* HASH_OPEN_MAP_HPP_ */
