#ifndef HASH_OPEN_SET_HPP_
#define HASH_OPEN_SET_HPP_

#include <string>
#include <iostream>
#include <sstream>
#include <initializer_list>
#include "ics_exceptions.hpp"
#include "pair.hpp"


namespace ics {


#ifndef undefinedhashdefined
#define undefinedhashdefined

    template<class T>
    int undefinedhash(const T &a) { return 0; }

#endif /* undefinedhashdefined */

//Instantiate the templated class supplying thash(a): produces a hash value for a.
//If thash is defaulted to undefinedhash in the template, then a constructor must supply chash.
//If both thash and chash are supplied, then they must be the same (by ==) function.
//If neither is supplied, or both are supplied but different, TemplateFunctionError is raised.
//The (unique) non-undefinedhash value supplied by thash/chash is stored in the instance variable hash.
    template<class T, int (*thash)(const T &a) = undefinedhash<T>>
    class HashOpenSet {
    public:
        typedef ics::pair<int, T> HashedElement;   //Stores hash(element) and element see the h_element member
        typedef int (*hashfunc)(const T &a);

        //Destructor/Constructors
        ~HashOpenSet();

        HashOpenSet(double the_load_threshold = 1.0, int (*chash)(const T &a) = undefinedhash<T>);

        explicit HashOpenSet(int initial_bins, double the_load_threshold = 1.0,
                             int (*chash)(const T &k) = undefinedhash<T>);

        HashOpenSet(const HashOpenSet<T, thash> &to_copy, double the_load_threshold = 1.0,
                    int (*chash)(const T &a) = undefinedhash<T>);

        explicit HashOpenSet(const std::initializer_list<T> &il, double the_load_threshold = 1.0,
                             int (*chash)(const T &a) = undefinedhash<T>);

        //Iterable class must support "for-each" loop: .begin()/.end() and prefix ++ on returned result
        template<class Iterable>
        explicit
        HashOpenSet(const Iterable &i, double the_load_threshold = 1.0, int (*chash)(const T &a) = undefinedhash<T>);


        //Queries
        bool empty() const;

        int size() const;

        bool contains(const T &element) const;

        std::string str() const; //supplies useful debugging information; contrast to operator <<

        //Iterable class must support "for-each" loop: .begin()/.end() and prefix ++ on returned result
        template<class Iterable>
        bool contains_all(const Iterable &i) const;


        //Commands
        int insert(const T &element);

        int erase(const T &element);

        void clear();

        //Iterable class must support "for" loop: .begin()/.end() and prefix ++ on returned result

        template<class Iterable>
        int insert_all(const Iterable &i);

        template<class Iterable>
        int erase_all(const Iterable &i);

        template<class Iterable>
        int retain_all(const Iterable &i);


        //Operators
        HashOpenSet<T, thash> &operator=(const HashOpenSet<T, thash> &rhs);

        bool operator==(const HashOpenSet<T, thash> &rhs) const;

        bool operator!=(const HashOpenSet<T, thash> &rhs) const;

        bool operator<=(const HashOpenSet<T, thash> &rhs) const;

        bool operator<(const HashOpenSet<T, thash> &rhs) const;

        bool operator>=(const HashOpenSet<T, thash> &rhs) const;

        bool operator>(const HashOpenSet<T, thash> &rhs) const;

        template<class T2, int (*hash2)(const T2 &a)>
        friend std::ostream &operator<<(std::ostream &outs, const HashOpenSet<T2, hash2> &s);


    public:
        class Iterator {
        public:
            //Private constructor called in begin/end, which are friends of HashOpenSet<T,thash>
            ~Iterator();

            T erase();

            std::string str() const;

            HashOpenSet<T, thash>::Iterator &operator++();

            HashOpenSet<T, thash>::Iterator operator++(int);

            bool operator==(const HashOpenSet<T, thash>::Iterator &rhs) const;

            bool operator!=(const HashOpenSet<T, thash>::Iterator &rhs) const;

            T &operator*() const;

            T *operator->() const;

            friend std::ostream &operator<<(std::ostream &outs, const HashOpenSet<T, thash>::Iterator &i) {
              outs << i.str(); //Use the same meaning as the debugging .str() method
              return outs;
            }

            friend Iterator HashOpenSet<T, thash>::begin() const;

            friend Iterator HashOpenSet<T, thash>::end() const;

        private:
            //If can_erase is false, current indexes the "next" value (must ++ to reach it)
            int current; //h_element Index
            HashOpenSet<T, thash> *ref_set;
            int expected_mod_count;
            bool can_erase = true;

            //Called in friends begin/end
            Iterator(HashOpenSet<T, thash> *iterate_over, int initial);
        };


        Iterator begin() const;

        Iterator end() const;


    private:
        int (*hash)(const T &k);          //Hashing function used (from template or constructor)

        const static int bin_empty = -1;   //Special values that can be stored in the bin array,
        const static int was_occupied = -2; //  along with non-negative indexes into the h_element array
        int *bin = nullptr; //Non-contiguous int array: bin[i] is either -1, -2, or an index into the h_element array
        HashedElement *h_element = nullptr; //Contiguous HashedElement array storing hash(element) and element as pair
        //bin/h_element are allocated in constructors and in ensure_lengths
        double load_threshold;              //used/bins must always be <= load_threshold (see ensure_lengths)
        int bins = 1;           //Allocated length of bin array (>= 1, so hash_compress doesn't % 0)
        int elements = 1;           //Allocated length of h_element array (can be < bins, if load_factor < 1)
        int used = 0;           //# of elements in this set; used <= elements (see ensure_lengths)
        //  bins-used = # of bin_empty/was_occupied values in bin array
        int mod_count = 0;           //For sensing concurrent modification


        //Helper methods
        //Given element, compute/return its hashed value and which bin it compresses to (%bins in range [0,bins-1])
        void hash_compress(const T &element, int &hashed, int &bin_index) const;

        int find_element(const T &element) const; //Returns index to element's bin or -1 if  element is not in any bin
        int next_unoccupied(int bin_index) const;   //this bin_index (if unoccupied) or next bin_index beyond that
        //   is unoccupied; advances using linear probing
        void ensure_lengths(
                int new_used);            //Reallocate bin/h_element if load factor > load_threshold/more elements needed
    };





//HashOpenSet class and related definitions

////////////////////////////////////////////////////////////////////////////////
//
//Destructor/Constructors

    template<class T, int (*thash)(const T &a)>
    HashOpenSet<T, thash>::~HashOpenSet() {
    }


    template<class T, int (*thash)(const T &a)>
    HashOpenSet<T, thash>::HashOpenSet(double the_load_threshold, int (*chash)(const T &element)) {
    }


    template<class T, int (*thash)(const T &a)>
    HashOpenSet<T, thash>::HashOpenSet(int initial_bins, double the_load_threshold, int (*chash)(const T &element)) {
    }

    template<class T, int (*thash)(const T &a)>
    HashOpenSet<T, thash>::HashOpenSet(const HashOpenSet<T, thash> &to_copy, double the_load_threshold,
                                       int (*chash)(const T &element)) {
    }


    template<class T, int (*thash)(const T &a)>
    HashOpenSet<T, thash>::HashOpenSet(const std::initializer_list<T> &il, double the_load_threshold,
                                       int (*chash)(const T &element)) {
    }


    template<class T, int (*thash)(const T &a)>
    template<class Iterable>
    HashOpenSet<T, thash>::HashOpenSet(const Iterable &i, double the_load_threshold, int (*chash)(const T &a)) {
    }

////////////////////////////////////////////////////////////////////////////////
//
//Queries

    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::empty() const {return false;
    }


    template<class T, int (*thash)(const T &a)>
    int HashOpenSet<T, thash>::size() const {return 0;
    }


    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::contains(const T &element) const {return false;
    }


    template<class T, int (*thash)(const T &a)>
    std::string HashOpenSet<T, thash>::str() const {return "";
    }


    template<class T, int (*thash)(const T &a)>
    template<class Iterable>
    bool HashOpenSet<T, thash>::contains_all(const Iterable &i) const {return false;
    }


////////////////////////////////////////////////////////////////////////////////
//
//Commands

    template<class T, int (*thash)(const T &a)>
    int HashOpenSet<T, thash>::insert(const T &element) {return 0;
    }


    template<class T, int (*thash)(const T &a)>
    int HashOpenSet<T, thash>::erase(const T &element) {return 0;
    }


    template<class T, int (*thash)(const T &a)>
    void HashOpenSet<T, thash>::clear() {
    }


    template<class T, int (*thash)(const T &a)>
    template<class Iterable>
    int HashOpenSet<T, thash>::insert_all(const Iterable &i) {return 0;
    }


    template<class T, int (*thash)(const T &a)>
    template<class Iterable>
    int HashOpenSet<T, thash>::erase_all(const Iterable &i) {return 0;
    }


    template<class T, int (*thash)(const T &a)>
    template<class Iterable>
    int HashOpenSet<T, thash>::retain_all(const Iterable &i) {return 0;
    }


////////////////////////////////////////////////////////////////////////////////
//
//Operators

    template<class T, int (*thash)(const T &a)>
    HashOpenSet<T, thash> &HashOpenSet<T, thash>::operator=(const HashOpenSet<T, thash> &rhs) {return *this;
    }


    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::operator==(const HashOpenSet<T, thash> &rhs) const {return false;
    }


    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::operator!=(const HashOpenSet<T, thash> &rhs) const {return false;
      return !(*this == rhs);
    }


    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::operator<=(const HashOpenSet<T, thash> &rhs) const {return false;
    }

    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::operator<(const HashOpenSet<T, thash> &rhs) const {return false;
    }


    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::operator>=(const HashOpenSet<T, thash> &rhs) const {return false;
    }


    template<class T, int (*thash)(const T &a)>
    bool HashOpenSet<T, thash>::operator>(const HashOpenSet<T, thash> &rhs) const {return false;
    }


    template<class T, int (*thash)(const T &a)>
    std::ostream &operator<<(std::ostream &outs, const HashOpenSet<T, thash> &s) {return outs;
    }


////////////////////////////////////////////////////////////////////////////////
//
//Iterator constructors

    template<class T, int (*thash)(const T &a)>
    auto HashOpenSet<T, thash>::begin() const -> HashOpenSet<T, thash>::Iterator {return Iterator(const_cast<HashOpenSet<T,thash>*>(this),0);
    }


    template<class T, int (*thash)(const T &a)>
    auto HashOpenSet<T, thash>::end() const -> HashOpenSet<T, thash>::Iterator {return Iterator(const_cast<HashOpenSet<T,thash>*>(this),0);
    }


////////////////////////////////////////////////////////////////////////////////
//
//Private helper methods

template<class T, int (*thash)(const T &a)>
void HashOpenSet<T, thash>::hash_compress(const T &element, int &hashed, int &bin_index) const {
}


template<class T, int (*thash)(const T &a)>
int HashOpenSet<T, thash>::find_element(const T &element) const {return 0;
}

template<class T, int (*thash)(const T &a)>
int HashOpenSet<T, thash>::next_unoccupied(int bin_index) const {return 0;
}


template<class T, int (*thash)(const T &a)>
void HashOpenSet<T, thash>::ensure_lengths(int new_used) {
}




////////////////////////////////////////////////////////////////////////////////
//
//Iterator class definitions

template<class T, int (*thash)(const T& a)>
HashOpenSet<T,thash>::Iterator::Iterator(HashOpenSet<T,thash>* iterate_over, int initial) {
}


template<class T, int (*thash)(const T& a)>
HashOpenSet<T,thash>::Iterator::~Iterator()
{}


template<class T, int (*thash)(const T& a)>
T HashOpenSet<T,thash>::Iterator::erase() {return ref_set->h_element[0].second;
}


template<class T, int (*thash)(const T& a)>
std::string HashOpenSet<T,thash>::Iterator::str() const {return "";
}


template<class T, int (*thash)(const T& a)>
auto  HashOpenSet<T,thash>::Iterator::operator ++ () -> HashOpenSet<T,thash>::Iterator& {return *this;
}


template<class T, int (*thash)(const T& a)>
auto  HashOpenSet<T,thash>::Iterator::operator ++ (int) -> HashOpenSet<T,thash>::Iterator {return *this;
}


template<class T, int (*thash)(const T& a)>
bool HashOpenSet<T,thash>::Iterator::operator == (const HashOpenSet<T,thash>::Iterator& rhs) const {return false;
}


template<class T, int (*thash)(const T& a)>
bool HashOpenSet<T,thash>::Iterator::operator != (const HashOpenSet<T,thash>::Iterator& rhs) const {return false;
}

template<class T, int (*thash)(const T& a)>
T& HashOpenSet<T,thash>::Iterator::operator *() const {return ref_set->h_element[0].second;
}

template<class T, int (*thash)(const T& a)>
T* HashOpenSet<T,thash>::Iterator::operator ->() const {return &(ref_set->h_element[0].second);
}

}

#endif /* HASH_OPEN_SET_HPP_ */
